import { ParamCalculability, ParamValueMode, Session } from "../../src/index";
import { MacroRugo, MacroRugoFlowType } from "../../src/macrorugo/macrorugo";
import { MacrorugoParams } from "../../src/macrorugo/macrorugo_params";
import { checkResult } from "../test_func";

/*
*** Emergent conditions Cd=1.5***
ks = 0.010000
D = 0.500000
k = 0.700000
Cd0 = 1.500000
S = 0.050000
B = 1.000000
h = 0.600000
C = 0.130000
RESULTS:
find_Q_nat(0.901710)=0.679969
Q=0.493729  fVal=0.000000
cote_bas = 12.200000
Vdeb = 0.822881
Fr = 0.530425
P = 242.174003
flowcond = emergent
Vmax = 1.799502
V_technique = 1.991299
q_technique = 0.561860
Strickler = 5.173107
 */

function macroRugoInstanceEmergentCd15(): MacroRugo {
    return new MacroRugo(
        new MacrorugoParams(
            12.5,   // ZF1
            6,      // L
            1,      // B
            0.05,   // If
            0.493729,    // Q
            0.6,    // h
            0.01,   // Ks
            0.13,   // C
            0.5,    // D
            0.7,    // k
            1.5     // Cd0
        )
    );
}

const macroRugoExtraResultEmergentCd15: { [key: string]: number|MacroRugoFlowType } = {
    ENUM_MacroRugoFlowType: MacroRugoFlowType.EMERGENT,
    Fr: 0.530425,
    PV: 242.174003,
    Vdeb: 0.822881,
    Vmax: 1.799502,
    Strickler: 5.173107,
    ZF2: 12.2
};

/*
*** Emergent conditions Cd=2***
ks = 0.010000
D = 0.500000
k = 0.700000
Cd0 = 2.000000
S = 0.050000
B = 1.000000
h = 0.600000
C = 0.130000
RESULTS:
find_Q_nat(0.768677)=0.653110
Q=0.376811  fVal=0.000000
cote_bas = 12.200000
Vdeb = 0.628018
Fr = 0.404817
P = 184.825731
flowcond = emergent
Vmax = 1.536129
V_technique = 1.592932
q_technique = 0.414154
Strickler = 3.948084
 */

function macroRugoInstanceEmergentCd2(): MacroRugo {
    const nub: MacroRugo = macroRugoInstanceEmergentCd15();
    nub.prms.Cd0.singleValue = 2;
    nub.prms.Q.singleValue = 0.376811;
    return nub;
}

const macroRugoExtraResultEmergentCd2: { [key: string]: number|MacroRugoFlowType } = {
    ENUM_MacroRugoFlowType: MacroRugoFlowType.EMERGENT,
    Fr: 0.404817,
    PV: 184.825731,
    Vdeb: 0.628018,
    Vmax: 1.536129,
    Strickler: 3.948084,
    ZF2: 12.2
};

/*
*** Submerged conditions ***
ks = 0.010000
D = 0.500000
k = 0.700000
Cd0 = 1.500000
S = 0.050000
B = 1.000000
h = 0.800000
C = 0.130000
RESULTS:
find_Q_nat(1.202280)=0.112522
Q=1.112263  fVal=0.000000
cote_bas = 12.200000
Vdeb = 1.390328
Fr = 0.776130
P = 545.564782
flowcond = immerge
q_technique = 0.940450
Strickler = 7.215044
 */

function macroRugoInstanceSubmerged(): MacroRugo {
    const nub: MacroRugo = macroRugoInstanceEmergentCd15();
    nub.prms.Y.singleValue = 0.8;
    nub.prms.Q.singleValue = 1.112263;
    return nub;
}

const macroRugoExtraResultSubmerged: { [key: string]: number|MacroRugoFlowType } = {
    ENUM_MacroRugoFlowType: MacroRugoFlowType.SUBMERGED,
    Fr: 0.776130,
    PV: 545.564782,
    Vdeb: 1.390328,
    Strickler: 7.215044,
    ZF2: 12.2
};

function MacroRugoFactory(sInstance: string): MacroRugo {
    let nub: MacroRugo;
    switch (sInstance) {
        case "EmergentCd15": {
            nub = macroRugoInstanceEmergentCd15();
            break;
        }
        case "EmergentCd2": {
            nub = macroRugoInstanceEmergentCd2();
            break;
        }
        case "Submerged": {
            nub = macroRugoInstanceSubmerged();
            break;
        }
        default: {
            throw new Error("Instance name error");
        }
    }
    for (const p of nub.parameterIterator) {
        p.v = p.singleValue;
    }
    return nub;
}

function testMacroRugo(sInstance: string, varTest: string, valRef: number) {
        it(`Calc(${varTest}) should be ${valRef}`, () => {
            const nub = MacroRugoFactory(sInstance);
            nub.prms.Q.v = nub.Calc("Q").vCalc;
            const p = nub.getParameter(varTest);
            p.v = undefined;
            p.valueMode = ParamValueMode.CALCUL;
            const r = nub.Calc(varTest, 0.1);
            checkResult(r, valRef);
        });
}

function testMacroRugoConfig(sInstance: string, Q0: number, fVal0: number, mrExtraRes: { [key: string]: number }) {
    describe(`Condition ${sInstance}` , () => {
        it(`resolveQ(${Q0}) should be ${fVal0}`, () => {
            const nubit = MacroRugoFactory(sInstance);
            nubit.prms.Q.v = Q0;
            // tslint:disable-next-line:no-string-literal
            nubit["setFlowType"]();
            // tslint:disable-next-line:no-string-literal
            expect(nubit["resolveQ"]()).toBeCloseTo(fVal0, 5);
        });
        const nub = MacroRugoFactory(sInstance);
        for (const prm of nub.prms) {
            if ([ParamCalculability.DICHO, ParamCalculability.EQUATION].includes(prm.calculability)) {
                testMacroRugo(sInstance, prm.symbol, prm.v);
            }
        }
        for (const sExtraRes in mrExtraRes) {
            if (mrExtraRes.hasOwnProperty(sExtraRes)) {
                it(`${sExtraRes} should be ${mrExtraRes[sExtraRes]}`, () => {
                    expect(MacroRugoFactory(sInstance).Calc("Q").values[sExtraRes])
                    .toBeCloseTo(mrExtraRes[sExtraRes], 5);
                });
            }
        }
    });
}

function macroRugoInstanceJalHyd85(): MacroRugo {
    const nubMR = macroRugoInstanceEmergentCd15();
    nubMR.prms.C.singleValue = 0.2;
    nubMR.prms.PBH.singleValue = 0.8;
    nubMR.calculatedParam = nubMR.prms.Q;
    return nubMR;
}

describe("Class MacroRugo: ", () => {

    testMacroRugoConfig("EmergentCd15", 0.901710, 0.679969, macroRugoExtraResultEmergentCd15);
    testMacroRugoConfig("EmergentCd2", 0.768677, 0.653110, macroRugoExtraResultEmergentCd2);
    // Le test passe en debug mais pas sous Jasmine !??
    // describe(`Condition Submerged` , () => {
    //     it(`resolveAlpha_t(0.07) should be 0.074752`, () => {
    //         // tslint:disable-next-line:no-string-literal
    //         expect(macroRugoInstanceSubmerged()["resolveAlpha_t"](0.07)).toBeCloseTo(0.074752, 5);
    //     });
    // });
    testMacroRugoConfig("Submerged", 1.202280, 0.112522, macroRugoExtraResultSubmerged);

    describe("JalHyd #85", () => {
        it("CalcSerie Q should return the good result :)", () => {
            const nubMR = macroRugoInstanceJalHyd85();
            const nubMR2 = macroRugoInstanceJalHyd85();

            nubMR.prms.Y.setValues(0.7, 1.2, 0.1);
            const aQ: number[] = [ 0.484255, 0.578363, 1.004229, 1.280269, 1.618937, 2.018687 ];

            nubMR.CalcSerie();
            for (let i = 0; i < aQ.length; i++) {
                expect(nubMR.result.resultElements[i].vCalc).toBeCloseTo(aQ[i], 4);
                nubMR2.prms.Y.setValue(0.7 + i * 0.1);
                expect(nubMR2.CalcSerie().vCalc).toBeCloseTo(aQ[i], 4);
            }
        });
    });
});
