import { ParallelStructure } from "./parallel_structure";

// Classes générales sur les structures
import { RectangularStructureParams } from "./rectangular_structure_params";
import { Structure } from "./structure";
import { LoiDebit } from "./structure_props";

// Equations de débit
import { StructureGateCem88d } from "./structure_gate_cem88d";
import { StructureGateCem88v } from "./structure_gate_cem88v";
import { StructureGateCunge80 } from "./structure_gate_cunge80";
import { StructureKivi } from "./structure_kivi";
import { StructureKiviParams } from "./structure_kivi_params";
import { StructureOrificeFree } from "./structure_orifice_free";
import { StructureOrificeFreeParams } from "./structure_orifice_free_params";
import { StructureOrificeSubmerged } from "./structure_orifice_submerged";
import { StructureOrificeSubmergedParams } from "./structure_orifice_submerged_params";
import { StructureRectangularOrificeFree } from "./structure_rectangular_orifice_free";
import { StructureRectangularOrificeSubmerged } from "./structure_rectangular_orifice_submerged";
import { StructureTriangularTruncWeirFree } from "./structure_triangular_trunc_weir";
import { TriangularTruncStructureParams } from "./structure_triangular_trunc_weir_params";
import { StructureTriangularWeir } from "./structure_triangular_weir";
import { StructureTriangularWeirBroad } from "./structure_triangular_weir_broad";
import { TriangularStructureParams } from "./structure_triangular_weir_params";
import { StructureVanLevLarinier } from "./structure_vanlev_larinier";
import { StructureVanLevParams } from "./structure_vanlev_params";
import { StructureVanLevVillemonte } from "./structure_vanlev_villemonte";
import { StructureWeirCem88d } from "./structure_weir_cem88d";
import { StructureWeirCem88v } from "./structure_weir_cem88v";
import { StructureWeirCunge80 } from "./structure_weir_cunge80";
import { StructureWeirFree } from "./structure_weir_free";
import { StructureWeirSubmerged } from "./structure_weir_submerged";
import { StructureWeirSubmergedLarinier } from "./structure_weir_submerged_larinier";
import { StructureWeirVillemonte } from "./structure_weir_villemonte";

export function CreateStructure(loiDebit: LoiDebit, parentNub?: ParallelStructure,
                                dbg: boolean = false
): Structure {
    let ret: Structure;
    const oCd: { [s: string]: number } = { SeuilR: 0.4, VanneR: 0.6, VanneRS: 0.8, SeuilT: 1.36 };
    const rectStructPrms: RectangularStructureParams = new RectangularStructureParams(
        0,  // Q
        100,        // ZDV
        102,        // Z1
        101.5,      // Z2
        2,          // L
        oCd.SeuilR, // Cd de seuil par défaut
        0.5 // W (pour les seuils réglé automatiquement à infinity par w.visible = false)
    );
    const vanLevPrms: StructureVanLevParams = new StructureVanLevParams(
        0,
        100,
        102,
        101.5,
        0.4,
        0.4,
        0.5,
        99.5,
        100.5
    );

    // Cd pour une vanne rectangulaire
    rectStructPrms.CdGR.singleValue = oCd.VanneR;
    rectStructPrms.CdGRS.singleValue = oCd.VanneRS;
    // Instanciation of the equation
    switch (loiDebit) {
        case LoiDebit.WeirCem88d:
            ret = new StructureWeirCem88d(rectStructPrms, dbg);
            break;

        case LoiDebit.WeirCem88v:
            ret = new StructureWeirCem88v(rectStructPrms, dbg);
            break;

        case LoiDebit.GateCem88d:
            ret = new StructureGateCem88d(rectStructPrms, dbg);
            break;

        case LoiDebit.GateCem88v:
            ret = new StructureGateCem88v(rectStructPrms, dbg);
            break;

        case LoiDebit.GateCunge80:
            ret = new StructureGateCunge80(rectStructPrms, dbg);
            rectStructPrms.CdCunge.singleValue = 1;
            break;

        case LoiDebit.WeirCunge80:
            ret = new StructureWeirCunge80(rectStructPrms, dbg);
            rectStructPrms.CdCunge.singleValue = 1;
            break;

        case LoiDebit.RectangularOrificeFree:
            ret = new StructureRectangularOrificeFree(rectStructPrms, dbg);
            break;

        case LoiDebit.RectangularOrificeSubmerged:
            ret = new StructureRectangularOrificeSubmerged(rectStructPrms, dbg);
            break;

        case LoiDebit.WeirFree:
            ret = new StructureWeirFree(rectStructPrms, dbg);
            break;

        case LoiDebit.WeirSubmergedLarinier:
            rectStructPrms.L.singleValue = 0.2;
            rectStructPrms.CdWSL.singleValue = 0.75;
            rectStructPrms.ZDV.singleValue = 101;
            rectStructPrms.h1.singleValue = 1;
            ret = new StructureWeirSubmergedLarinier(rectStructPrms, dbg);
            break;

        case LoiDebit.WeirSubmerged:
            rectStructPrms.CdWS.singleValue = 0.9;
            ret = new StructureWeirSubmerged(rectStructPrms, dbg);
            break;

        case LoiDebit.KIVI:
            const structKiviPrm: StructureKiviParams = new StructureKiviParams(
                8.516,      // Q
                101.5,        // ZDV
                103,        // Z1
                102,        // Z2
                2,          // L
                oCd.SeuilR, // alpha
                0.001,      // béta
                100);       // ZRAM : cote Radier Amont
            structKiviPrm.h1.singleValue = 1.5;
            ret = new StructureKivi(structKiviPrm, dbg);
            break;

        case LoiDebit.TriangularWeirFree:
        case LoiDebit.TriangularWeirBroad:
            const structTriangPrms: TriangularStructureParams = new TriangularStructureParams(
                0,          // Q
                100,        // ZDV
                102,        // Z1
                100,        // Z2
                45,         // Alpha2
                oCd.SeuilT  // Cd pour un seuil triangulaire
                // W = Infinity par défaut pour un seuil
            );
            if(loiDebit === LoiDebit.TriangularWeirFree) {
                ret = new StructureTriangularWeir(structTriangPrms, dbg);
            } else {
                ret = new StructureTriangularWeirBroad(structTriangPrms, dbg);
            }
            break;

        case LoiDebit.TriangularTruncWeirFree:
            const structTriTruncPrms: TriangularTruncStructureParams = new TriangularTruncStructureParams(
                0,          // Q
                100.1,        // ZDV
                102,        // Z1
                100.1,        // Z2
                0.9,        // BT
                101,        // ZT
                oCd.SeuilT  // Cd pour un seuil triangulaire
                // W = Infinity par défaut pour un seuil
            );
            ret = new StructureTriangularTruncWeirFree(structTriTruncPrms, dbg);
            break;

        case LoiDebit.OrificeSubmerged:
            ret = new StructureOrificeSubmerged(
                new StructureOrificeSubmergedParams(
                    0,      // Q
                    102,    // Z1
                    101.5,  // Z2
                    0.7,    // Cd
                    0.1     // S
                ),
                dbg
            );
            break;

        case LoiDebit.OrificeFree:
            ret = new StructureOrificeFree(
                new StructureOrificeFreeParams(
                    0,      // Q
                    102,    // Z1
                    101.5,  // Z2
                    0.7,    // Cd
                    0.1,    // S
                    101     // Zco
                ),
                dbg
            );
            break;

        case LoiDebit.WeirVillemonte:
            ret = new StructureWeirVillemonte(rectStructPrms, dbg);
            break;

        case LoiDebit.VanLevVillemonte:
            ret = new StructureVanLevVillemonte(vanLevPrms, dbg);
            break;

        case LoiDebit.VanLevLarinier:
            vanLevPrms.CdWSL.singleValue = 0.75;
            ret = new StructureVanLevLarinier(vanLevPrms, dbg);
            break;

        default:
            throw new Error(`type de LoiDebit ${LoiDebit[loiDebit]} non pris en charge`);
    }

    // set reference to parent
    if (parentNub) {
        ret.parent = parentNub;
    }

    return ret;
}
